# -*- coding: utf-8 -*-
"""

.. moduleauthor:: Vitaley Zaretskey <vitaley@whisper.sh>

"""

from __future__ import division, print_function

import numpy as np
import pandas as pd

from numpy.random import choice


def bootstrap_ci(data, statistic=np.mean, alpha=0.1, name=None, n_draws=10000):
    """
    calculate bootstrap estimated confidence interval for the provided statistic

    :param data: the data, 1-d numpy array or anything else that can be converted into 1-d array
    :param statistic: the statistic to calculate, can be any callable that accepts a 1-d numpy array
        and returns a single orderable value
    :param alpha: 1 - alpha is the width of the confidence interval
    :param name: optional, name of the statistic to prepend output labels
    :param n_draws: number of bootstrap draws to take to estimate the confidence interval
    :return: pandas Series with lower, center, and upper estimates of the statistic
    """

    prefix = name + '_' if name else ''
    n = len(data)
    data = np.array(data).ravel()
    idx = int(n_draws * alpha / 2)

    b_stats = np.zeros(shape=n_draws, dtype=np.float64)
    for i in range(n_draws):
        b_draw = choice(data, size=n, replace=True)
        b_stats[i] = statistic(b_draw)

    b_stats.partition(kth=[idx-1, -idx])
    center = statistic(data)
    lower = b_stats[idx-1]
    upper = b_stats[-idx]

    return pd.Series({
        prefix + 'lower': lower,
        prefix + 'center': center,
        prefix + 'upper': upper})
