import re

from . import Preprocessor

class NontrivialPreprocessor(Preprocessor): 
    def __init__(self, config):
        self._config = config
        self._contraction_expansion_regex = self._build_substitution_regex(self._config.contraction_expansion_dict)
        self._d_to_a_regex = self._build_substitution_regex(self._config.d_to_a_dict)
        self._space_punctuation_regex = re.compile("([{0:s}])".format("".join(map(re.escape, self._config.puncts))))
        self._contract_emoticons_regex = self._build_substitution_regex(self._config.emoticons_dict)
        self._final_corrections_regex = self._build_substitution_regex(self._config.final_corrections_dict)

    def preprocess(self, text):
        if isinstance(text, basestring):
            return self._preprocess(text)
        elif isinstance(text, (list, tuple)):
            return map(self._preprocess, text)

    def _preprocess(self, text): 
        """
        step 1: lowercase 
        step 2: expand contractions 
        step 3: spell out numbers 
        step 4: separate out emoticons/emojis and normalize punctuation 
        step 5: correct stuff the previous steps mess up (ex. < three -> <3 (it really matters!))
        step 6: reduce whitespaces
        """
        text = text.lower()
        text = self._expand_contractions(text)
        text = self._spell_out_numbers(text)
        text = self._normalize_punctuation(text)
        text = self._correct(text)
        return self._reduce_whitespaces(text)

    def _build_substitution_regex(self, substitution_dict): 
        if len(substitution_dict) > 0:
            return re.compile("(%s)" % "|".join(map(re.escape, substitution_dict.keys())))
        return re.compile('a^') # this will match nothing, which we want if the contractions dict is empty

    def _substitute(self, substitution_dict, regex, text):
        return regex.sub(lambda mo: substitution_dict[mo.string[mo.start():mo.end()]], text)

    def _reduce_whitespaces(self, text): 
        return re.sub("\s+", " ", text).strip()

    def _move_apostrophes(self, text): 
        return re.sub(r"'", " '", text)
    
    def _expand_contractions(self, text): 
        # expand any contractions
        return self._move_apostrophes(self._substitute(self._config.contraction_expansion_dict, self._contraction_expansion_regex, text))

    def _spell_out_numbers(self, text): 
        return self._substitute(self._config.d_to_a_dict, self._d_to_a_regex, text)

    def _space_punctuation(self, text):
        return self._space_punctuation_regex.sub(r" \1 ", text)

    def _contract_emoticons(self, text):
        return self._substitute(self._config.emoticons_dict, self._contract_emoticons_regex, self._reduce_whitespaces(text) + " ")

    def _space_emojis(self, text): 
        return self._config.emojis_regex.sub(r" \1 ", text)

    def _normalize_punctuation(self, text): 
        text = self._space_punctuation(text)
        text = self._contract_emoticons(text)
        text = self._space_emojis(text)
        text = re.sub(r"\?\s+(\?\s*)+", " ??? ", text)  # any more than one question mark goes to just one 
        text = re.sub(r"\.\s+(\.\s*)+", " ... ", text)  # convert any occurence of >1 periods to ellipsis
        text = re.sub(r"\!\s+(\!\s*)+", " !!! ", text)  # ditto for exclamation marks 
        return text

    def _correct(self, text): 
        return self._substitute(self._config.final_corrections_dict, self._final_corrections_regex, text)
